@@ -1,4 +1,4 @@
-# Copyright: 2015-2016 Tim Harder <radhermit@gmail.com>
+# Copyright: 2015-2017 Tim Harder <radhermit@gmail.com>
 # Copyright: 2008-2011 Brian Harring <ferringb@gmail.com>
 # License: BSD/GPL2
 
@@ -10,11 +10,12 @@ passing in distutils.
 Specifically, this module is only meant to be imported in setup.py scripts.
 """
 
+from contextlib import contextmanager
 import copy
 import errno
-import inspect
 import io
 import math
+from multiprocessing import cpu_count
 import operator
 import os
 import re
@@ -27,6 +28,7 @@ import textwrap
 os.environ["SNAKEOIL_DEMANDLOAD_PROTECTION"] = 'n'
 os.environ["SNAKEOIL_DEMANDLOAD_WARN"] = 'n'
 
+from setuptools import find_packages
 from setuptools.command import install as dst_install
 
 from distutils import log
@@ -40,47 +42,73 @@ from distutils.command import (
 READTHEDOCS = os.environ.get('READTHEDOCS', None) == 'True'
 
 # top level repo/tarball directory
-TOPDIR = os.path.dirname(os.path.abspath(inspect.stack(0)[1][1]))
+TOPDIR = os.path.dirname(os.path.abspath(__file__))
 
 
-def find_project(topdir=TOPDIR):
-    """Determine a project's name.
+def find_moduledir(searchdir=TOPDIR):
+    """Determine a module's directory path.
 
     Based on the assumption that the project is only distributing one main
     module.
     """
-    topdir_depth = len(topdir.split('/'))
     modules = []
+    moduledir = None
+    searchdir_depth = len(searchdir.split('/'))
+
+    # allow modules to be found inside a top-level src dir
+    if os.path.exists(os.path.join(TOPDIR, 'src')):
+        searchdir_depth += 1
 
     # look for a top-level module
-    for root, dirs, files in os.walk(topdir):
-        # only descend at most one level
-        if len(root.split('/')) > topdir_depth + 1:
+    for root, dirs, files in os.walk(searchdir):
+        # only descend to a specified level
+        if len(root.split('/')) > searchdir_depth + 1:
             continue
         if '__init__.py' in files:
-            modules.append(os.path.basename(root))
+            modules.append(root)
 
-    if not modules:
-        raise ValueError('No project module found')
+    if len(modules) == 1:
+        moduledir = modules[0]
     elif len(modules) > 1:
-        raise ValueError('Multiple project modules found: %s' % (', '.join(modules)))
+        # Multiple modules found in the base directory, searching for one that
+        # defines __title__.
+        main_modules = []
+        for path in modules:
+            with io.open(os.path.join(path, '__init__.py'), encoding='utf-8') as f:
+                try:
+                    main_modules.append(re.search(
+                        r'^__title__\s*=\s*[\'"]([^\'"]*)[\'"]',
+                        f.read(), re.MULTILINE).group(1))
+                except AttributeError:
+                    continue
 
-    return modules[0]
+        if not main_modules or len(main_modules) > 1:
+            raise ValueError(
+                'Multiple main modules found in %r: %s' % (
+                    searchdir, ', '.join(os.path.basename(x) for x in modules)))
+        else:
+            moduledir = modules[0]
+
+    if moduledir is None:
+        raise ValueError('No main module found')
+
+    return moduledir
 
 
-# determine the project we're being imported into
-PROJECT = find_project()
+# determine the main module we're being used to package
+MODULEDIR = find_moduledir()
+PACKAGEDIR = os.path.dirname(MODULEDIR)
+MODULE = os.path.basename(MODULEDIR)
 
 
-def version(project=PROJECT):
-    """Determine a project's version.
+def version(moduledir=MODULEDIR):
+    """Determine a module's version.
 
-    Based on the assumption that a project defines __version__ in its main
-    module.
+    Based on the assumption that a module defines __version__.
     """
     version = None
     try:
-        with io.open(os.path.join(TOPDIR, project, '__init__.py'), encoding='utf-8') as f:
+        with io.open(os.path.join(moduledir, '__init__.py'), encoding='utf-8') as f:
             version = re.search(
                 r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]',
                 f.read(), re.MULTILINE).group(1)
@@ -91,16 +119,16 @@ def version(project=PROJECT):
             raise
 
     if version is None:
-        raise RuntimeError('Cannot find version for project: %s' % (project,))
+        raise RuntimeError('Cannot find version for module: %s' % (MODULE,))
 
     return version
 
 
-def readme(project=PROJECT):
+def readme(topdir=TOPDIR):
     """Determine a project's long description."""
     for doc in ('README.rst', 'README'):
         try:
-            with io.open(os.path.join(TOPDIR, doc), encoding='utf-8') as f:
+            with io.open(os.path.join(topdir, doc), encoding='utf-8') as f:
                 return f.read()
         except IOError as e:
             if e.errno == errno.ENOENT:
@@ -111,6 +139,67 @@ def readme(project=PROJECT):
     return None
 
 
+def setup():
+    """Parameters and commands for setuptools."""
+    params = {
+        'name': MODULE,
+        'version': version(),
+        'long_description': readme(),
+        'packages': find_packages(PACKAGEDIR),
+        'package_dir': {'':os.path.basename(PACKAGEDIR)},
+        'install_requires': install_requires(),
+    }
+
+    cmds = {
+        'sdist': sdist,
+    }
+
+    # check for scripts
+    if os.path.exists(os.path.join(TOPDIR, 'bin')):
+        params['scripts'] = os.listdir('bin')
+        cmds['build_scripts'] = build_scripts
+
+    # check for docs
+    if os.path.exists(os.path.join(TOPDIR, 'doc')):
+        cmds['build_docs'] = build_docs
+        cmds['install_docs'] = install_docs
+
+    # check for man pages
+    if os.path.exists(os.path.join(TOPDIR, 'doc', 'man')):
+        cmds['build_man'] = build_man
+        cmds['install_man'] = install_man
+
+    return params, cmds
+
+
+def _requires(path):
+    """Determine a project's various dependencies from requirements files."""
+    try:
+        with io.open(path) as f:
+            return f.read().splitlines()
+    except IOError as e:
+        if e.errno == errno.ENOENT:
+            pass
+        else:
+            raise
+    return None
+
+
+def build_requires():
+    """Determine a project's build dependencies."""
+    return _requires(os.path.join(TOPDIR, 'requirements', 'build.txt'))
+
+
+def install_requires():
+    """Determine a project's runtime dependencies."""
+    return _requires(os.path.join(TOPDIR, 'requirements', 'install.txt'))
+
+
+def test_requires():
+    """Determine a project's test dependencies."""
+    return _requires(os.path.join(TOPDIR, 'requirements', 'test.txt'))
+
+
 def get_file_paths(path):
     """Get list of all file paths under a given path."""
     for root, dirs, files in os.walk(path):
@@ -169,6 +258,42 @@ def pkg_config(*packages, **kw):
     return kw
 
 
+def cython_pyx(path=MODULEDIR):
+    """Return all available cython extensions under a given path."""
+    for root, _dirs, files in os.walk(path):
+        for f in files:
+            if f.endswith('.pyx'):
+                yield str(os.path.join(root, f))
+
+
+def cython_exts(build_deps=None, build_exts=None, build_opts=None, path=MODULEDIR):
+    """Prepare all cython extensions under a given path to be built."""
+    build_deps = build_deps if build_deps is not None else []
+    build_exts = build_exts if build_exts is not None else []
+    build_opts = build_opts if build_opts is not None else {}
+
+    exts = []
+    cython_exts = []
+
+    for ext in cython_pyx(path):
+        cythonized = os.path.splitext(ext)[0] + '.c'
+        if os.path.exists(cythonized):
+            exts.append(cythonized)
+        else:
+            cython_exts.append(ext)
+
+    # require cython to build exts as necessary
+    if cython_exts:
+        build_deps.append('cython')
+        exts.extend(cython_exts)
+
+    build_exts.extend(
+        Extension(os.path.splitext(ext)[0].replace(os.path.sep, '.'), [ext], **build_opts)
+        for ext in exts)
+
+    return build_deps, build_exts
+
+
 class OptionalExtension(Extension):
     """Python extension that is optional to build.
 
@@ -183,8 +308,6 @@ class OptionalExtension(Extension):
 class sdist(dst_sdist.sdist):
     """sdist command wrapper to bundle generated files for release."""
 
-    package_namespace = PROJECT
-
     def initialize_options(self):
         dst_sdist.sdist.initialize_options(self)
 
@@ -194,12 +317,13 @@ class sdist(dst_sdist.sdist):
         This is used by the --version option in interactive programs among
         other things.
         """
-        from snakeoil.version import get_git_version
+        with syspath(PACKAGEDIR, MODULE == 'snakeoil'):
+            from snakeoil.version import get_git_version
         log.info('generating _verinfo')
         data = get_git_version(base_dir)
         if not data:
             return
-        path = os.path.join(base_dir, self.package_namespace, '_verinfo.py')
+        path = os.path.join(base_dir, '_verinfo.py')
         with open(path, 'w') as f:
             f.write('version_info=%r' % (data,))
 
@@ -211,25 +335,30 @@ class sdist(dst_sdist.sdist):
         exist in a working tree.
         """
 
-        if 'build_man' in self.distribution.cmdclass:
+        # don't build man pages when running under tox
+        if ('build_man' in self.distribution.cmdclass and
+                not os.path.basename(os.environ.get('_', '')) == 'tox'):
+            build_man = self.reinitialize_command('build_man')
+            build_man.ensure_finalized()
             self.run_command('build_man')
-            shutil.copytree(os.path.join(os.getcwd(), "build/sphinx/man"),
-                            os.path.join(base_dir, "man"))
+            shutil.copytree(os.path.join(os.getcwd(), build_man.content_search_path[0]),
+                            os.path.join(base_dir, build_man.content_search_path[1]))
 
         dst_sdist.sdist.make_release_tree(self, base_dir, files)
-        self.generate_verinfo(base_dir)
+        build_py = self.reinitialize_command('build_py')
+        build_py.ensure_finalized()
+        self.generate_verinfo(os.path.join(
+            base_dir, build_py.package_dir.get('', ''), MODULE))
 
     def run(self):
         build_ext = self.reinitialize_command('build_ext')
         build_ext.ensure_finalized()
 
         # generate cython extensions if any exist
-        cython = any(
-            os.path.splitext(f)[1] == '.pyx' for e in
-            build_ext.extensions for f in e.sources)
-        if cython:
+        extensions = list(cython_pyx())
+        if extensions:
             from Cython.Build import cythonize
-            cythonize(build_ext.extensions)
+            cythonize(extensions, nthreads=cpu_count())
 
         dst_sdist.sdist.run(self)
 
@@ -240,7 +369,6 @@ class build_py(dst_build_py.build_py):
     user_options = dst_build_py.build_py.user_options + \
         [("inplace", "i", "do any source conversions in place")]
 
-    package_namespace = PROJECT
     generate_verinfo = True
 
     def initialize_options(self):
@@ -255,10 +383,11 @@ class build_py(dst_build_py.build_py):
 
     def _run_generate_verinfo(self, rebuilds=None):
         ver_path = self.get_module_outfile(
-            self.build_lib, (self.package_namespace,), '_verinfo')
+            self.build_lib, (MODULE,), '_verinfo')
         # this should check mtime...
         if not os.path.exists(ver_path):
-            from snakeoil.version import get_git_version
+            with syspath(PACKAGEDIR, MODULE == 'snakeoil'):
+                from snakeoil.version import get_git_version
             log.info('generating _verinfo')
             with open(ver_path, 'w') as f:
                 f.write("version_info=%r" % (get_git_version('.'),))
@@ -283,7 +412,9 @@ class build_py2to3(build_py):
             except EnvironmentError:
                 # ok... wtf distutils?
                 continue
-            trg_path = os.path.join(self.build_lib, path)
+            trg_path = os.path.join(
+                self.build_lib,
+                path.lstrip(self.package_dir.get('', '')).lstrip(os.path.sep))
             if force:
                 yield trg_path, new_mtime
                 continue
@@ -300,11 +431,11 @@ class build_py2to3(build_py):
 
     def get_py2to3_converter(self, options=None, proc_count=0):
         from lib2to3 import refactor as ref_mod
-        from snakeoil.dist import caching_2to3
+        with syspath(PACKAGEDIR, MODULE == 'snakeoil'):
+            from snakeoil.dist import caching_2to3
 
         if proc_count == 0:
-            import multiprocessing
-            proc_count = multiprocessing.cpu_count()
+            proc_count = cpu_count()
 
         assert proc_count >= 1
 
@@ -383,6 +514,18 @@ class build_py3to2(build_py2to3):
         log.info("completed py2k conversions")
 
 
+def generate_html():
+    """Generate html docs for the project."""
+    from snakeoil.dist.generate_docs import generate_html
+    generate_html(TOPDIR, PACKAGEDIR, MODULE)
+
+
+def generate_man():
+    """Generate man pages for the project."""
+    from snakeoil.dist.generate_docs import generate_man
+    generate_man(TOPDIR, PACKAGEDIR, MODULE)
+
+
 class build_man(Command):
     """Build man pages.
 
@@ -391,6 +534,7 @@ class build_man(Command):
     2to3 or other conversions instead of straight from the build directory.
     """
 
+    description = "build man pages"
     user_options = [
         ("force", "f", "force build as needed"),
     ]
@@ -417,25 +561,22 @@ class build_man(Command):
             build_py = self.reinitialize_command('build_py')
             build_py.ensure_finalized()
             self.run_command('build_py')
-            syspath = sys.path[:]
-            sys.path.insert(0, os.path.abspath(build_py.build_lib))
+            with syspath(os.path.abspath(build_py.build_lib)):
+                # generate man page content for scripts we create
+                if 'build_scripts' in self.distribution.cmdclass:
+                    generate_man()
 
-            # generate man page content for scripts we create
-            if 'build_scripts' in self.distribution.cmdclass:
-                from snakeoil.dist.generate_docs import generate_man
-                generate_man(PROJECT, TOPDIR)
-
-            # generate man pages
-            build_sphinx = self.reinitialize_command('build_sphinx')
-            build_sphinx.builder = 'man'
-            build_sphinx.ensure_finalized()
-            self.run_command('build_sphinx')
-            sys.path = syspath
+                # generate man pages
+                build_sphinx = self.reinitialize_command('build_sphinx')
+                build_sphinx.builder = 'man'
+                build_sphinx.ensure_finalized()
+                self.run_command('build_sphinx')
 
 
 class build_docs(build_man):
     """Build html docs."""
 
+    description = "build HTML documentation"
     user_options = [
         ("force", "f", "force build as needed"),
     ]
@@ -453,8 +594,7 @@ class build_docs(build_man):
             self.run_command('build_man')
 
             # generate API docs
-            from snakeoil.dist.generate_docs import generate_html
-            generate_html(PROJECT, TOPDIR)
+            generate_html()
 
             # generate html docs -- allow build_sphinx cmd to run again
             build_sphinx = self.reinitialize_command('build_sphinx')
@@ -545,7 +685,7 @@ class build_ext(dst_build_ext.build_ext):
             any(not os.path.exists(x) for ext in self.no_cythonize() for x in ext.sources))
         if use_cython:
             from Cython.Build import cythonize
-            cythonize(self.extensions)
+            cythonize(self.extensions, nthreads=cpu_count())
 
         self.extensions = self.no_cythonize()
         return dst_build_ext.build_ext.run(self)
@@ -588,8 +728,8 @@ class build_scripts(dst_build_scripts.build_scripts):
                     #!%s
                     from os.path import basename
                     from %s import scripts
-                    scripts.main(basename(__file__))
-                """ % (sys.executable, PROJECT)))
+                    scripts.run(basename(__file__))
+                """ % (sys.executable, MODULE)))
         self.copy_scripts()
 
 
@@ -629,6 +769,7 @@ class install_docs(Command):
     """Install html documentation."""
 
     content_search_path = build_docs.content_search_path
+    description = "install HTML documentation"
     user_options = [
         ('path=', None, "final path to install to; else it's calculated"),
         ('build-dir=', None, "build directory"),
@@ -657,7 +798,7 @@ class install_docs(Command):
 
     def calculate_install_path(self):
         return os.path.join(
-            os.path.abspath(self.prefix), 'share', 'doc', PROJECT + '-%s' % version(), 'html')
+            os.path.abspath(self.prefix), 'share', 'doc', MODULE + '-%s' % version(), 'html')
 
     def find_content(self):
         for possible_path in self.content_search_path:
@@ -711,6 +852,7 @@ class install_docs(Command):
 class install_man(install_docs):
     """Install man pages."""
 
+    description = "install man pages"
     content_search_path = build_man.content_search_path
     build_command = 'build_man'
 
@@ -765,6 +907,7 @@ class test(Command):
 
     blacklist = frozenset()
 
+    description = "run unit tests in a built copy"
     user_options = [
         ("inplace", "i", "do building/testing in place"),
         ("skip-rebuilding", "s", "skip rebuilds. primarily for development"),
@@ -777,7 +920,7 @@ class test(Command):
         ("include-dirs=", "I", "include dirs for build_ext if needed"),
     ]
 
-    default_test_namespace = '%s.test' % PROJECT
+    default_test_namespace = '%s.test' % MODULE
 
     def initialize_options(self):
         self.inplace = False
@@ -839,7 +982,7 @@ class test(Command):
 
         # remove temporary plugincache so it isn't installed
         plugincache = os.path.join(
-            os.path.abspath(build_py.build_lib), build_py.package_namespace,
+            os.path.abspath(build_py.build_lib), MODULE,
             'plugins/plugincache')
         if os.path.exists(plugincache):
             os.remove(plugincache)
@@ -848,39 +991,60 @@ class test(Command):
             raise DistutilsExecError("tests failed; return %i" % (retval,))
 
 
-class PyTest(Command):
-    """Run tests using pytest against a built copy."""
+class pytest(Command):
+    """Run tests using pytest."""
 
+    description = "run unit tests in a built copy using pytest"
     user_options = [
         ('pytest-args=', 'a', 'arguments to pass to py.test'),
         ('coverage', 'c', 'generate coverage info'),
+        ('skip-build', 's', 'skip building the module'),
+        ('test-dir=', 'd', 'directory to source tests from'),
         ('report=', 'r', 'generate and/or show a coverage report'),
         ('jobs=', 'j', 'run X parallel tests at once'),
         ('match=', 'k', 'run only tests that match the provided expressions'),
     ]
 
-    default_test_dir = os.path.join(TOPDIR, PROJECT, 'test')
-
     def initialize_options(self):
         self.pytest_args = ''
         self.coverage = False
+        self.skip_build = False
+        self.test_dir = None
         self.match = None
         self.jobs = None
         self.report = None
 
     def finalize_options(self):
-        self.test_args = [self.default_test_dir]
+        # if a test dir isn't specified explicitly try to find one
+        if self.test_dir is None:
+            for path in (os.path.join(TOPDIR, 'test'),
+                         os.path.join(TOPDIR, 'tests'),
+                         os.path.join(MODULEDIR, 'test'),
+                         os.path.join(MODULEDIR, 'tests')):
+                if os.path.exists(path):
+                    self.test_dir = path
+                    break
+            else:
+                raise DistutilsExecError('cannot automatically determine test directory')
+
+        self.test_args = [self.test_dir]
         self.coverage = bool(self.coverage)
+        self.skip_build = bool(self.skip_build)
+        if self.verbose:
+            self.test_args.append('-v')
         if self.match is not None:
             self.test_args.extend(['-k', self.match])
 
-        if self.coverage:
+        if self.coverage or self.report:
             try:
                 import pytest_cov
-                self.test_args.extend(['--cov', PROJECT])
+                self.test_args.extend(['--cov', MODULE])
             except ImportError:
-                sys.stderr.write('error: install pytest-cov for coverage support\n')
-                sys.exit(1)
+                raise DistutilsExecError('install pytest-cov for coverage support')
+
+            coveragerc = os.path.join(TOPDIR, '.coveragerc')
+            if os.path.exists(coveragerc):
+                self.test_args.extend(['--cov-config', coveragerc])
 
             if self.report is None:
                 # disable coverage report output
@@ -893,8 +1057,7 @@ class PyTest(Command):
                 import xdist
                 self.test_args.extend(['-n', self.jobs])
             except ImportError:
-                sys.stderr.write('error: install pytest-xdist for -j/--jobs support\n')
-                sys.exit(1)
+                raise DistutilsExecError('install pytest-xdist for -j/--jobs support')
 
         # add custom pytest args
         self.test_args.extend(shlex.split(self.pytest_args))
@@ -903,27 +1066,61 @@ class PyTest(Command):
         try:
             import pytest
         except ImportError:
-            sys.stderr.write('error: pytest is not installed\n')
-            sys.exit(1)
+            raise DistutilsExecError('pytest is not installed')
 
-        # build extensions and byte-compile python
-        build_ext = self.reinitialize_command('build_ext')
-        build_py = self.reinitialize_command('build_py')
-        build_ext.ensure_finalized()
-        build_py.ensure_finalized()
-        self.run_command('build_ext')
-        self.run_command('build_py')
+        if self.skip_build:
+            builddir = MODULEDIR
+        else:
+            # build extensions and byte-compile python
+            build_ext = self.reinitialize_command('build_ext')
+            build_py = self.reinitialize_command('build_py')
+            build_ext.ensure_finalized()
+            build_py.ensure_finalized()
+            self.run_command('build_ext')
+            self.run_command('build_py')
+            builddir = os.path.abspath(build_py.build_lib)
 
-        # Change the current working directory to the builddir during testing
-        # so coverage paths are correct.
-        builddir = os.path.abspath(build_py.build_lib)
-        if self.coverage and os.path.exists(os.path.join(TOPDIR, '.coveragerc')):
-            shutil.copyfile(os.path.join(TOPDIR, '.coveragerc'),
-                            os.path.join(builddir, '.coveragerc'))
-        ret = subprocess.call([sys.executable, '-m', 'pytest'] + self.test_args, cwd=builddir)
+        with syspath(builddir):
+            from snakeoil.contexts import chdir
+            # Change the current working directory to the builddir during testing
+            # so coverage paths are correct.
+            with chdir(builddir):
+                ret = pytest.main(self.test_args)
         sys.exit(ret)
 
 
+class pylint(Command):
+    """Run pylint on a module."""
+
+    description = "run pylint on a module"
+    user_options = [
+        ('errors-only', 'E', 'Check only errors with pylint'),
+        ('output-format=', 'f', 'Change the output format'),
+    ]
+
+    def initialize_options(self):
+        self.errors_only = False
+        self.output_format = 'colorized'
+
+    def finalize_options(self):
+        self.errors_only = bool(self.errors_only)
+
+    def run(self):
+        try:
+            from pylint import lint
+        except ImportError:
+            raise DistutilsExecError('pylint is not installed')
+
+        lint_args = [MODULEDIR]
+        rcfile = os.path.join(TOPDIR, '.pylintrc')
+        if os.path.exists(rcfile):
+            lint_args.extend(['--rcfile', rcfile])
+        if self.errors_only:
+            lint_args.append('-E')
+        lint_args.extend(['--output-format', self.output_format])
+        lint.Run(lint_args)
+
+
 def print_check(message, if_yes='found', if_no='not found'):
     """Decorator to print pre/post-check messages."""
     def sub_decorator(f):
@@ -994,7 +1191,8 @@ class config(dst_config.config):
         return self.try_link("int main(int argc, char *argv[]) { return 0; }")
 
     def run(self):
-        from snakeoil.pickling import dump, load
+        with syspath(PACKAGEDIR, MODULE == 'snakeoil'):
+            from snakeoil.pickling import dump, load
 
         # try to load the cached results
         try:
@@ -1042,6 +1240,27 @@ class config(dst_config.config):
             % (typename, member), headers, include_dirs, lang)
 
 
+# directly copied from snakeoil.contexts
+@contextmanager
+def syspath(path, condition=True, position=0):
+    """Context manager that mangles sys.path and then reverts on exit.
+
+    Args:
+        path: The directory path to add to sys.path.
+        condition: Optional boolean that decides whether sys.path is mangled or
+            not, defaults to being enabled.
+        position: Optional integer that is the place where the path is inserted
+            in sys.path, defaults to prepending.
+    """
+    syspath = sys.path[:]
+    if condition:
+        sys.path.insert(position, path)
+    try:
+        yield
+    finally:
+        sys.path = syspath
+
+
 # yes these are in snakeoil.compatibility; we can't rely on that module however
 # since snakeoil source is in 2k form, but this module is 2k/3k compatible.
 # in other words, it could be invoked by py3k to translate snakeoil to py3k
